/*******************************************************************************

Program:

	Main.cc
	01.00.00

Title:

	Basic Translater Main Function

Description:

	This is the main driver function for the basic translater.

Algorithm:

	This translater operates in steps on the source code:

	1. Rough translation using LEX and YACC to create a tree out
	   of the entire program. This tree only approximates the actual
	   program sequence.

	2. Scan through tree, removing out initilization code, and making
	   tables of line numbers and variables used in the program.

	3. Write out the variable definition section of the output.

	4. Modify the tree to actual execution format, converting
	   if-then-else, for-next, ... into a better node format.

	5. Lose line numbers and labels that are not used.

	6. Pull subroutines out into seperate units.

	7. Optimize tree. See if we can replace statements with
		auto-increment/decrement, ...

	8. Write out code section of program.

Bugs:

	Requires the source code to be formatted into a more strict
	format than the actual basic compiler likes.

History:

	01.01.00 - 08/30/1996 - Kevin Handy
		Set version number to 1.1, and install on internet.

	01.01.01 - 08/31/1996 - Kevin Handy
		Changed 'TreeStruct' to 'Node', which makes more sense.
		Modified for privatization of Nodes data.
		Lose extra calls to 'UnDownLink', which werent needed.
		Found bug in "IF x THEN IF y THEN z else a else b"

	01.01.02 - 10/22/1996 - Kevin Handy
		Added variables 'IntegerType' and 'RealType'.

	01.02.01 - 12/17/1996 - Kevin Handy
		Move initalization of 'IntegerType' and 'RealType' from
		parse.y to main().

	01.02.02 - 01/05/1997 - Kevin Handy
		Created a function to write out the titles.
		Added date and time of translation.

*******************************************************************************/

//
// System Include Files
//
#include <iostream.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <assert.h>
#include <time.h>

//
// Project Include Files
//
#include "basic.h"
#include "nodes.h"
#include "parse.h"

//
// Module #defined Constants
//

//
// Module #defined Functions
//

//
// Global Variables
//
VariableList *Variables;

int CompileFlag;

int NeedIostreamH = 0;
int NeedStringH = 0;
int NeedMathH = 0;
int NeedErrorH = 0;
int NeedStdlibH = 0;
int NeedRFA = 0;
int NeedBasicFun = 0;
int NeedTimeH = 0;
int NeedDataList = 0;

int IntegerType = VARTYPE_INTEGER;
int RealType = VARTYPE_REAL;

int DebugDump = FALSE;
int DebugDumpOne = FALSE;
int VariableDump = FALSE;
int PositionDump = FALSE;


//
// Module Structure Definitions
//

//
// Module Function Prototypes
//
static void WriteHeader(ostream& os, const char* Name);

//
// Module Variables
//
static ostream *OutFile;

/*******************************************************************************

Global Function:

	main()

Title:

	Main driver function.

Description:

	This is the overall controlling function of the translater.

Bugs:

	Needs command line options.

*******************************************************************************/


main(int argc, char *argv[])
{
	//
	// Local Variables
	//
	int ArgLoop;
	int DidOneFlag = 0;

	//
	// Initialize variables
	//
	OutFile = &cout;
	CompileFlag = 0;

	//
	// Handle input arguements
	//
	for (ArgLoop = 1; ArgLoop < argc; ArgLoop++)
	{
		if (argv[ArgLoop][0] == '-')
		{
			switch(tolower(argv[ArgLoop][1]))
			{
			case 't':
				//
				// Tracing
				//
				switch(tolower(argv[ArgLoop][2]))
				{
				case '1':
					DebugDump = TRUE;
					break;
				case '2':
					DebugDumpOne = TRUE;
					break;
				default:
					DebugDump = TRUE;
					DebugDumpOne = TRUE;
					break;
				}
				break;

			case 'v':
				//
				// Variable dump(s)
				//
				VariableDump = TRUE;
				break;

			case 'o':
				//
				// Output File
				//
				if (OutFile != &cout)
				{
					delete OutFile;
					OutFile = NULL;
				}

				ArgLoop++;

				OutFile = new ofstream(argv[ArgLoop]);
				if (OutFile == NULL)
				{
					cerr << "Unable to open " << argv[ArgLoop] <<
						"for output" << endl;
					exit(1);
				}

				break;

			case 'h':
				DidOneFlag = 1;
				cerr << "Usage: basic [-t] [-v] [-c] [-p] "
					"[-o output] source" << endl;
				break;

			case 'c':
				CompileFlag = 1;
				break;

			case 'p':
				PositionDump = TRUE;
				break;

			default:
				cerr << "%Error on command line " <<
					argv[ArgLoop] << endl;
				break;
			}
		}
		else
		{
			//
			// At least one file name on command line
			//
			DidOneFlag = 1;

			//
			// Reset line counter
			//
			xline = 1;
			include_stack_pointer = 0;

			//
			// Open up source file
			//
			if ((yyin = fopen(argv[ArgLoop], "r")) == 0)
			{
				cerr << "Unable to open input file " <<
					argv[ArgLoop] << endl;
				exit(1);
			}

			WriteHeader(*OutFile, argv[ArgLoop]);

//			*OutFile << "//" << endl <<
//				"// " << argv[ArgLoop] << endl <<
//				"// Translated from Basic to C++ using btran(~tm)" <<
//				endl <<
//				"//" << endl << endl;

			//
			// Step 1. Parse the program. YACC will call the
			//  rest of the program if it scans correctly.
			//
			if (PositionDump)
			{
				cerr << "Parsing." << endl;
			}

			Variables = new VariableList;

			/*
			 * Defaulvariable size
			 */
			IntegerType = VARTYPE_INTEGER;
			RealType = VARTYPE_REAL;

			if (yyparse() != 0)
			{
				cerr << "%Failure during parse" << endl;
				delete Variables;
				fclose(yyin);
				exit(1);
			}

			//
			// Close the input file
			//
			fclose(yyin);
			delete Variables;
		}
	}

	if (DidOneFlag == 0)
	{
		//
		// Reset line counter
		//
		xline = 1;
		include_stack_pointer = 0;

		//
		// Read from standard input if we haven't handled
		// any input files yet.
		//
		yyin = stdin;

		WriteHeader(*OutFile, "Standard Input");

//		*OutFile << "//" << endl <<
//			"// Translated from Basic to C++ using btran(~tm)" <<
//			endl <<
//			"//" << endl << endl;

		if (PositionDump)
		{
			cerr << "Parsing." << endl;
		}

		Variables = new VariableList;

		if (yyparse() != 0)
		{
			cerr << "%Failure during parse" << endl;
			delete Variables;
			exit(1);
		}

		//
		// Close the input file
		//
		delete Variables;
	}

	if (OutFile != &cout)
	{
		delete OutFile;
	}

	return 0;
}

/*******************************************************************************

Local function:

	WriteHeader()

Title:

	Write Header To Output File

Description:

	Writes out a nice little title at the start if an output
	translation saying I'll take the blame for the lousy code
	output.
*******************************************************************************/

static void WriteHeader(ostream& os, const char* Name)
{
	time_t timer;
	struct tm *tblock;
	char Text[64];

	timer = time(NULL);
	tblock = localtime(&timer);

	strftime(Text, sizeof(Text), "%A, %B %d, %Y at %X",
		tblock);

	os << "//" << endl;
	os << "// Source: " << Name << endl;
	os << "// Translated from Basic to C++ using btran(~tm:)" << endl;
	os << "// on " << Text << endl;
	os << "//" << endl << endl;
}


/*******************************************************************************

Global Function:

	DoProgram()

Title:

	Main part of translater

Description:

	This function does most of the work after the parser has
	generated the tree.

Bugs:

	Not yet complete.

*******************************************************************************/

void DoProgram(Node *Program)
{

	//
	// Pretend it all started with a FUNCTION INTGER MAIN.
	// This creates a 'main' function for the program.
	//
	Node* BeginProgram = new Node(BAS_S_FUNCTION);
	Node* BeginType = new Node(BAS_S_INTEGER);
	Node* BeginName = new Node(BAS_V_FUNCTION, "MAIN");

	BeginProgram->Link(BeginType, NULL, BeginName);
	BeginProgram->DownLink(Program);

	//
	// Dump out accourding to flags
	//
	if (DebugDump && (Program != 0))
	{
		BeginProgram->PrintTree();
	}

	if (PositionDump)
	{
		cerr << "Scanning for loops" << endl;
	}
	BeginProgram = ScanForLoops(BeginProgram);

	if (DebugDumpOne)
	{
		BeginProgram->PrintTree();
	}

	/*
	 * Step 8. Output the translated code
	 */
	if (PositionDump)
	{
		cerr << "Outputing code" << endl;
	}
	BeginProgram->Output(*OutFile);

	//
	// Free up all memory allocated
	//
	delete BeginProgram;
}

/*******************************************************************************

Global Function:

	ScanForLoops()

Title:

	Fix up loops, if-then-else, functions.

Description:

	This function scans through the program, fixing up for-next loops,
	if-then-else statements, while statements, program segments, etc.
	Since the parser doesn't make complete loops for anything except
	modified statements, we need to roll the statements together
	here.

Bugs:

	Goes internally into the tree to restructure the code instead of
	using the interface code.

*******************************************************************************/

Node *ScanForLoops
(
	Node *Program
)
{
	//
	// Local Variables
	//
	Node *ThisTree = Program;
	Node *NextTree;
	Node *IndentTree[128];
	int IndentType[128];
	int ThisLevel = 0;

	//
	// Start the whole mass going
	//
	assert(Program != 0);
	IndentTree[0] = 0;

	//
	// Scan through the program, looking for loops at level 0.
	//
	while (ThisTree != 0)
	{
		//
		// Remember what the next level will be
		//
		NextTree = ThisTree->GetDown();
		ThisTree->UnDownLink();

		//
		// Preview this node, and decide what to do with it.
		//
		switch(ThisTree->GetType())
		{
		case BAS_S_THEN:
			//
			// We don't need to keep THEN's at all.
			//
//			ThisTree->UnDownLink();
			delete ThisTree;
			goto reloop;

		case BAS_S_ELSE:
			//
			// Look for matching IF
			//
			while ((ThisLevel > 0) &&
				((IndentType[ThisLevel-1] != BAS_S_IF) ||
				(IndentTree[ThisLevel-1]->GetTree(2) != NULL)))
			{
				ThisLevel--;
			}

			//
			// We better be in an if statement
			//
			if (IndentType[ThisLevel-1] == BAS_S_IF)
			{
				//
				// Use the ELSE node as a hook to this
				// section.
				//
				assert(IndentTree[ThisLevel-1]->GetTree(2) == NULL);
				IndentTree[ThisLevel-1]->Link(NULL, NULL, ThisTree);
				IndentTree[ThisLevel] = ThisTree;
			}
			else
			{
				cerr << "Missing IF for ELSE" << endl;
			}

			//
			// We don't need to keep the 'ELSE' any longer.
			//
//			ThisTree->UnDownLink();
//			delete ThisTree;
			goto reloop;

		case BAS_S_NEXT:
			//
			// Next statement
			//
			// Look for previous "for" or "while". This may cause
			// IF's to terminate early, but if the program is
			// structured correctly we will only go back one
			// level.
			//
			while ((ThisLevel > 0) &&
				(IndentType[ThisLevel-1] != BAS_S_FOR) &&
				(IndentType[ThisLevel-1] != BAS_S_WHILE) &&
				(IndentType[ThisLevel-1] != BAS_N_FORUNTIL) &&
				(IndentType[ThisLevel-1] != BAS_N_FORWHILE))
			{
				ThisLevel--;
			}

			//
			// Now back up one level
			//
			if (ThisLevel > 0)
			{
				ThisLevel--;
			}

			//
			// We don't need to keep NEXT any longer
			//
//			ThisTree->UnDownLink();
			delete ThisTree;
			goto reloop;

		case BAS_N_ENDRECORD:
			//
			// END RECORD statement
			//
			// Look for previous "RECORD".
			//
			while ((ThisLevel > 0) &&
				(IndentType[ThisLevel-1] != BAS_S_RECORD))
			{
				ThisLevel--;
			}

			//
			// Now back up one level
			//
			if (ThisLevel > 0)
			{
				ThisLevel--;
			}

			//
			// We don't need to keep 'END RECORD' any longer
			//
//			ThisTree->UnDownLink();
			delete ThisTree;
			goto reloop;

		case BAS_N_ENDVARIANT:
			//
			// END VARIANT statement
			//
			// Look for previous "VARIANT".
			//
			while ((ThisLevel > 0) &&
				(IndentType[ThisLevel-1] != BAS_S_VARIANT))
			{
				ThisLevel--;
			}

			//
			// Now back up one level
			//
			if (ThisLevel > 0)
			{
				ThisLevel--;
			}

			//
			// We don't need to keep 'END VARIANT' any longer
			//
//			ThisTree->UnDownLink();
			delete ThisTree;
			goto reloop;

		case BAS_S_FUNCTIONEND:
		case BAS_S_SUBEND:
			//
			// End a function/subroutine
			//
			// Force back to lowest level
			//
			ThisLevel = 0;

			//
			// We don't need to keep 'END SUB' any longer
			//
//			ThisTree->UnDownLink();
			delete ThisTree;
			goto reloop;

		case BAS_S_FNEND:
			//
			// End an internal function
			//
			while ((ThisLevel > 0) &&
				(IndentType[ThisLevel-1] != BAS_S_DEF))
			{
				ThisLevel--;
			}

			//
			// Now back up one level
			//
			if (ThisLevel > 0)
			{
				ThisLevel--;
			}

			//
			// We don't need to keep FNEND any longer
			//
//			ThisTree->UnDownLink();
			delete ThisTree;
			goto reloop;

		case BAS_V_LABEL:
			//
			// A line number will terminate any open IF
			// statements. I'll assume they are all at the
			// top of the stack.
			//
			if (isdigit(ThisTree->GetText()[0]))
			{
				while ((ThisLevel > 0) &&
					(IndentType[ThisLevel-1] == BAS_S_IF))
				{
					ThisLevel--;
				}
			}

			//
			// We need to keep line number.
			//
			break;

		case BAS_N_ENDIF:
			//
			// end-if
			//
			if ((ThisLevel > 0) &&
				(IndentType[ThisLevel-1] == BAS_S_IF))
			{
				ThisLevel--;
			}

			//
			// Lose the 'END IF' statement.
			//
//			ThisTree->UnDownLink();
			delete ThisTree;
			goto reloop;

		case BAS_S_CASE:
		case BAS_N_CASEELSE:
			while ((ThisLevel > 0) &&
				(IndentType[ThisLevel-1] != BAS_S_SELECT) &&
				(IndentType[ThisLevel-1] != BAS_S_VARIANT))
			{
				ThisLevel--;
			}

			break;

		case BAS_N_ENDSELECT:
			while ((ThisLevel > 0) &&
				(IndentType[ThisLevel-1] != BAS_S_SELECT))
			{
				ThisLevel--;
			}

			if (ThisLevel > 0)
			{
				ThisLevel--;
			}

			//
			// Lose the 'END SELECT'
			//
//			ThisTree->UnDownLink();
			delete ThisTree;
			goto reloop;

		case BAS_S_FUNCTION:
		case BAS_S_SUB:
			if (ThisTree->GetTree(3) == NULL)
			{
				//
				// A Function or Subroutine
				//
				ThisLevel = 0;
			}
			break;
		}

		//
		// Connect this node into the current level
		//
		if (IndentTree[ThisLevel] == 0)
		{
			//
			// If we start with a empty level,
			// we need to hook this
			// node down to the previous level.
			//
			if (ThisLevel != 0)
			{
				assert(IndentTree[ThisLevel-1]->GetTree(3) == NULL);
				IndentTree[ThisLevel-1]->Link(NULL, NULL, NULL, ThisTree);
			}
		}
		else
		{
			//
			// Link Nodes to the previous node on same level
			//
			IndentTree[ThisLevel]->DownLink(ThisTree);
		}

		IndentTree[ThisLevel] = ThisTree;

		//
		// Look at this node, and decide what to do with it
		//
		switch(ThisTree->GetType())
		{
		case BAS_S_IF:
		case BAS_S_UNLESS:
		case BAS_S_FUNCTION:
		case BAS_S_SUB:
		case BAS_S_FOR:
		case BAS_N_FORUNTIL:
		case BAS_N_FORWHILE:
		case BAS_S_UNTIL:
		case BAS_S_WHILE:
		case BAS_S_DEF:

			if (ThisTree->GetTree(3) == NULL)
			{
				IndentType[ThisLevel] = ThisTree->GetType();
				ThisLevel++;
				IndentTree[ThisLevel] = 0;
			}
			break;

		case BAS_S_CASE:
		case BAS_N_CASEELSE:
		case BAS_S_SELECT:
		case BAS_S_RECORD:
		case BAS_S_VARIANT:

			assert(ThisTree->GetTree(3) == NULL);

			IndentType[ThisLevel] = ThisTree->GetType();
			ThisLevel++;
			IndentTree[ThisLevel] = 0;
			break;
		}

		//
		// Point to next node
		//
reloop:
		ThisTree = NextTree;
	}

	//
	// Done. Hand back re-structured tree.
	//
	return(Program);
}

#ifndef __MSDOS__

/*******************************************************************************

Global Function:

	strupr()

Title:

	Convert a string to upper case

Description:

	This function will convert a text string to upper case.

Bugs:

	This function exists in MSDOS, but not in ANSI-C definition,
	and thus is defined here to make it available.

*******************************************************************************/

char *strupr(char *s)
{
	char *s1 = s;

	while (*s1)
	{
		*s1 = toupper(*s1);
		s1++;
	}

	return s;
}

/*******************************************************************************

Global Function:

	strlwr()

Title:

	Convert a string to lower case

Description:

	This function will convert a text string to lower case.

Bugs:

	This function exists in MSDOS, but not in ANSI-C definition,
	and thus is defined here to make it available.

*******************************************************************************/

char *strlwr(char *s)
{
	char *s1 = s;

	while (*s1)
	{
		*s1 = tolower(*s1);
		s1++;
	}

	return s;
}
#endif
