/*******************************************************************************

Program:

	nodes.cc
	01.00.00

Title:

	Routines to work with nodes.

Description:

	This file contains various routines for manipulating the tree
	nodes used when parsing a basic program.

Compile:

	cc [noopt]

History:

	01.01.00 - 08/30/1996 - Kevin Handy
		Update version number to 1.1, and install on internet.

	01.01.01 - 08/31/1996 - Kevin Handy
		Changed 'TreeStruct' to 'Node', which makes mode sense.

	01.01.02 - 09/19/1996 - Kevin Handy
		Always include 'bstring.h' for string variables.

	01.01.03 - 09/20/1996 - Kevin Handy
		Include <errno.h> instead of <error.h> (oops).
		Always include <stdlib.h>

	01.01.04 - 10/06/1996 - Kevin Handy
		Added BAS_V_PREDEF capability.

	01.01.05 - 10/20/1996 - Kevin Handy
		Put a semicolon ';' after labels. This eliminates the
		error "label must be followed by a statement" from G++.
		Output 'include "datalist.h"' when needed.

	01.02.01 - 11/11/1996 - Kevin Handy
		Added parameter to output functions (ostream& os),
		so that I didn't have to sneak in the channel through
		a side path. (Also loose weird *OutFile<< syntax)

	01.02.02 - 11/11/1996 - Kevin Handy
		Midifications to use "os <<" instead of calls to
		"Outputxxx(os)" whenever possible.

	01.02.03 - 11/12/1996 - Kevin Handy
		Modifications to reduce the number of functions,
		along with changes to parse.y

	01.02.04 - 12/27/1996 - Kevin Handy
		Make the INPUT statement do something again.

	01.02.05 - 01/05/1997 - Kevin Handy
		Force string constants to be BString's when necessary.

*******************************************************************************/

//
// System Include Files
//
#include <iostream.h>
#include <stdio.h>
#include <assert.h>
#include <stddef.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <assert.h>

//
// Project Include Files
//
#include "basic.h"
#include "parse.h"
#include "nodes.h"

//
// Module #defined Constants
//
#define MAXLEVEL 256                    // Maximum depth of nodes

//
// Module #defined Functions
//

//
// Module Structure Definitions
//

//
// Module Function Prototypes
//
static char *mangle_string(char *Text);
static void mangle_string_value(char *ThisPart, int Base, char Type);

//
// Module Variables
//

/*******************************************************************************

Member Function:

	Node()

Title:

	Constructs a node.

Description:

	This function creates a new node.

*******************************************************************************/

int Node::Level = 0;

Node::Node(int TypeValue, char* xTextValue, int Include)
{
	Type = TypeValue;
	TextValue = NULL;
	Down = NULL;
	Tree[0] = NULL;
	Tree[1] = NULL;
	Tree[2] = NULL;
	Tree[3] = NULL;
	FromInclude = Include;

	//
	// Allocate room for text, and copy over.
	//
	if (xTextValue != NULL)
	{
		char *ThisText = new char[strlen(xTextValue) + 1];
		assert(ThisText != NULL);
		strcpy(ThisText, xTextValue);
		if ((Type != BAS_V_TEXTSTRING) && (Type != BAS_S_REMARK))
		{
			strupr(ThisText);
		}
		TextValue = ThisText;
	}
}

/*******************************************************************************

Member Function:

	~Node()

Title:

	Destructs a node, and all of it's children.

Description:

	This function frees an existing node, and all of its children.

*******************************************************************************/

//
// Destructor
//
Node::~Node()
{
	// Lose all parameters

	delete Tree[0];
	delete Tree[1];
	delete Tree[2];
	delete Tree[3];

	// Lose text value

	delete TextValue;
}

/*******************************************************************************


	Link()

Title:

	Link child nodes into parent node.

Description:

	This function will link several child nodes into a parent node.

Return Value:

	Returns a pointer to the parent node (passed as the first
	parameter)

Bugs:

*******************************************************************************/

Node *Node::Link
(
	Node *Node0,
	Node *Node1,
	Node *Node2,
	Node *Node3
)
{
	//
	// Attach them together
	//
	if (Node0 != NULL)
	{
		assert(Tree[0] == NULL);
		assert(Node0 != this);
		Tree[0] = Node0;
	}
	if (Node1 != NULL)
	{
		assert(Tree[1] == NULL);
		assert(Node1 != this);
		Tree[1] = Node1;
	}
	if (Node2 != NULL)
	{
		assert(Tree[2] == NULL);
		assert(Node2 != this);
		Tree[2] = Node2;
	}
	if (Node3 != NULL)
	{
		assert(Tree[3] == NULL);
		assert(Node3 != this);
		Tree[3] = Node3;
	}

	//
	// Return main node back
	//
	return(this);
}

/*******************************************************************************

Member Function:

	DownLink()

Title:

	Link two nodes together at same level.

Description:

	This function will link two parent nodes together at the same
	tree level. If one or the other node doesn't exist, it will
	make whichever node exists the main node.

Return Value:

	Returns the top node.
Bugs:

*******************************************************************************/

//
// Link several nodes together through the Nodes
//
Node *Node::DownLink
(
	Node *Node0
)
{
	//
	// Don't link if nothing to link to
	//
	if (this == NULL)
	{
		return(Node0);
	}

	//
	// Skip out if link node is null
	//
	if (Node0 == NULL)
	{
		return this;
	}

	assert(Node0 != this);

	//
	// If we aren't at the top of the stack, scan to the end
	//
	Node* LookDown = this;
	while (LookDown->Down != NULL)
	{
		LookDown = LookDown->Down;
	}

	//
	// Attach them together
	//
	LookDown->Down = Node0;

	//
	// Return main node back
	//
	return(this);
}

/*******************************************************************************

Global Function:

	PrintTree()

Title:

	Prints the node tree.

Description:

	This function will print out a simple picture of the node
	and its relationship to it's children.

Bugs:

	Only displays to stdout. Ugly display.

*******************************************************************************/

//
// Print out tree in a tree structure
//
void Node::PrintTree(char* Level, int Flag, char Node)
{
	//
	// Not really a node
	//
	assert(this != NULL);

	//
	// Local variables
	//
	char* DownLevel;

	//
	// Make sure we have an active node
	//
	if ((Down != NULL) || Flag)
	{
		DownLevel = " | ";
	}
	else
	{
		DownLevel = "   ";
	}

	char* NextLevel = new char[strlen(Level) + 4];
	strcpy(NextLevel, Level);
	strcat(NextLevel, DownLevel);

	//
	// Print bars for this level
	//
	cout << Level << " " << Node << "-" << Type;

	if (TextValue != NULL)
	{
		cout << " Text: " << TextValue;
	}

	cout << endl;

	//
	// Print Parameter Levels
	//
	if (Tree[0] != NULL)
	{
		Tree[0]->PrintTree(NextLevel,
			(Tree[1] != NULL) | (Tree[2] != NULL) |
			(Tree[3] != NULL), '0');
	}

	if (Tree[1] != NULL)
	{
		Tree[1]->PrintTree(NextLevel,
			(Tree[2] != NULL) | (Tree[3] != NULL), '1');
	}
	if (Tree[2] != NULL)
	{
		Tree[2]->PrintTree(NextLevel, (Tree[3] != NULL), '2');
	}

	if (Tree[3] != NULL)
	{
		Tree[3]->PrintTree(NextLevel, 0, '3');
	}

	//
	// Link down to next parent if possible
	//
	if (Down != NULL)
	{
		Down->PrintTree(Level, 0, 'D');
	}
}

/*******************************************************************************

Member Function:

	VariableScan()

Title:

	Search block for variable names

Description:

	Scans a block of code for variable names.
	This must be done before scanning for the loops.
	Call this funstion to process a block of statements.

Bugs:

*******************************************************************************/

void Node::VariableScan(int InDefineFlag)
{
	Node *ThisNode;

	assert(this != NULL);

	//
	// Scan this statement
	//
	VariableScanOne(InDefineFlag);

	//
	// Scan Next Statement(s)
	//
	//	I'm doing the scan here instead of having each node
	//	call scan on it's own Dowm because the recursion becomes
	//	a real memory killer on large programs.
	//
	ThisNode = Down;

	while (ThisNode != NULL)
	{
		ThisNode->VariableScanOne(InDefineFlag);
		ThisNode = ThisNode->Down;
	}
}

/*******************************************************************************

Member Function:

	VariableScanOne()

Title:

	Search statement for variable names.

Description:

	Scans a block of code for variable names.
	This must be done before scanning for the loops.
	Call this function to process a single statement.

Bugs:

*******************************************************************************/

void Node::VariableScanOne(int InDefineFlag)
{
	VariableStruct *ThisVar;
	int ThisType = 0;

	assert(this != NULL);

	//
	// Look at current node type
	//
	switch (Type)
	{
	case BAS_N_EXTERNALCONSTANT:
		assert(Tree[2] != NULL);
		if (Tree[1] != NULL)
		{
			ThisType = Tree[1]->ScanType();
		}
		else
		{
			ThisType = 0;
		}

		Tree[2]->ScanVarList(ThisType, VARCLASS_CONST, TRUE);

		break;

	case BAS_N_EXTERNALFUNCTION:
		assert(Tree[1] != NULL);
		ThisType = Tree[1]->ScanType();
		Tree[2]->ScanVarList(ThisType, VARCLASS_FUNC, TRUE);
		break;

	case BAS_N_EXTERNALSUB:
		Tree[1]->ScanVarList(VARTYPE_VOID, VARCLASS_FUNC, TRUE);
		break;

	case BAS_S_CALL:
		Tree[0]->ScanVarList(VARTYPE_VOID, VARCLASS_FUNC, FALSE);
		break;

	case BAS_S_COMMON:
		assert(Tree[0] != NULL);

		ThisVar = new VariableStruct(Tree[0]->TextValue,
			VARTYPE_STRUCT, VARCLASS_NONE, TRUE);
		Variables->Append(ThisVar);
		Variables->Fixup();

		break;

	case BAS_S_DATA:
		break;

	case BAS_S_DECLARE:
		assert(Tree[2] != NULL);
		if (Tree[1] != NULL)
		{
			ThisType = Tree[1]->ScanType();
		}
		else
		{
			ThisType = 0;
		}

		Tree[2]->ScanVarList(ThisType, VARCLASS_CONST, TRUE);

		break;

	case BAS_S_DIM:
		assert(Tree[0] != NULL);
		Tree[0]->VariableScanOne(1);
		break;

	case BAS_S_EXTERNAL:
		assert(Tree[2] != NULL);
		if (Tree[1] != NULL)
		{
			ThisType = Tree[1]->ScanType();
		}
		else
		{
			ThisType = 0;
		}

		Tree[2]->ScanVarList(ThisType, VARCLASS_NONE, TRUE);

		break;

	case BAS_S_MAP:

		ScanMap();
		break;

	case BAS_V_FUNCTION:
		//
		// Built in function
		//
		ThisVar = Variables->Lookup(TextValue, this);
		if (ThisVar == NULL)
		{
			//
			// Create variable
			//
			ThisVar = new VariableStruct(TextValue,
				VARTYPE_VOID, VARCLASS_FUNC, InDefineFlag);
			assert(ThisVar != NULL);

			//
			// Add to variable table
			//
			Variables->Append(ThisVar);
		}
		else
		{
			ThisVar->SetClass(VARCLASS_FUNC);
		}
		if (Tree[0] != NULL)
		{
			Tree[0]->VariableScan(InDefineFlag);
		}
		if (Tree[3] != NULL)
		{
			Tree[3]->VariableScan(0);
		}
		break;

	case BAS_S_FUNCTION:
		//
		// Function Definition
		//
		assert(Tree[2] != NULL);

		if (Tree[0] != NULL)
		{
			ThisType = Tree[0]->ScanType();
		}
		else
		{
			ThisType = VARTYPE_VOID;
		}

		ThisVar = Variables->Lookup(Tree[2]->TextValue, this);
		if (ThisVar == NULL)
		{
			//
			// Create variable
			//
			ThisVar = new VariableStruct(Tree[2]->TextValue,
				ThisType, VARCLASS_FUNC, InDefineFlag);
			assert(ThisVar != NULL);

			//
			// Add to variable table
			//
			Variables->Append(ThisVar);
		}
		else
		{
			ThisVar->SetClass(VARCLASS_FUNC);
			ThisVar->Type = ThisType;
		}

		//
		// Scan calling parameters
		//
		if (Tree[2]->Tree[0] != NULL)
		{
			Tree[2]->Tree[0]->VariableScan(1);
		}

		//
		// Scan source code
		//
		if (Tree[3] != NULL)
		{
			Tree[3]->VariableScan(0);
		}

		break;

	case BAS_V_LABEL:
		ThisVar = Variables->Lookup(TextValue, 0);
		if (ThisVar == NULL)
		{
			int SeekColon = strlen(TextValue);
			if (TextValue[SeekColon - 1] == ':')
			{
				TextValue[SeekColon-1] = '\0';
			}
			ThisVar = new VariableStruct(TextValue,
				VARTYPE_LABEL, VARCLASS_LABEL, TRUE);
			assert(ThisVar != NULL);
			Variables->Append(ThisVar);
		}
		break;

	case BAS_V_USELABEL:
		ThisVar = Variables->Lookup(TextValue, 0);
		if (ThisVar == NULL)
		{
			ThisVar = new VariableStruct(TextValue,
				VARTYPE_LABEL, VARCLASS_LABEL, TRUE);
			assert(ThisVar != NULL);
			Variables->Append(ThisVar);
		}
		ThisVar->EverUsed++;
		break;

	case BAS_V_NAME:
		Variables->Define(this, FALSE, InDefineFlag);

		//
		// Look at possible array reference
		//
		if (Tree[0] != NULL)
		{
			Tree[0]->VariableScanOne(InDefineFlag);
		}

		if (Tree[2] != NULL)
		{
			Tree[2]->VariableScanOne(InDefineFlag);
		}

		break;

	case BAS_S_RECORD:
		assert(Tree[0] != NULL);
		assert(Tree[3] != NULL);
		Tree[0]->VariableScanOne(1);
		Tree[3]->VariableScan(1);
		break;

	case BAS_N_STRUCTNAME:

		//
		// Force structer variable to be usable under C++
		//
		strlwr(TextValue);

		//
		// Look at possible array reference
		//
		if (Tree[0] != NULL)
		{
			Tree[0]->VariableScanOne(InDefineFlag);
		}

		if (Tree[2] != NULL)
		{
			Tree[2]->VariableScanOne(InDefineFlag);
		}

		break;

	case BAS_X_STRREF:

		assert (Tree[0] != NULL);
		assert (Tree[1] != NULL);

		Tree[0]->VariableScanOne(InDefineFlag);
		Tree[1]->VariableScanOne(InDefineFlag);

		break;

	default:
		if (Tree[0] != NULL)
		{
			Tree[0]->VariableScan(InDefineFlag);
		}

		if (Tree[1] != NULL)
		{
			Tree[1]->VariableScan(InDefineFlag);
		}

		if (Tree[2] != NULL)
		{
			Tree[2]->VariableScan(InDefineFlag);
		}

		if (Tree[3] != NULL)
		{
			Tree[3]->VariableScan(InDefineFlag);
		}
		break;
	}
}


/*******************************************************************************

Member Function:

	Node::ScanVarList()

Title:

	Scans a (possible) list for the variable names of the
	current type.

Description:

	Scans a block of code for the variable type.

Bugs:

*******************************************************************************/

void Node::ScanVarList(int ThisType, int ThisClass, int Status)
{
	assert(this != NULL);

	switch(Type)
	{
	case BAS_N_LIST:
	case BAS_N_ASSIGNLIST:

		Tree[0]->ScanVarList(ThisType, ThisClass, Status);
		Tree[1]->ScanVarList(ThisType, ThisClass, Status);
		break;

	case '=':
	case BAS_N_ASSIGN:

		Tree[0]->ScanVarList(ThisType, ThisClass, Status);
		break;

	default:
		VariableStruct* ThisVar = Variables->Lookup(TextValue, Tree[0]);
		if (ThisVar == NULL)
		{
			ThisVar = new VariableStruct(TextValue,
				ThisType, ThisClass, Status);
			assert(ThisVar != NULL);
			Variables->Append(ThisVar);
		}
	}
}


/*******************************************************************************

Member Function:

	Node::ScanType()

Title:

	Search block for type definion

Description:

	Scans a block of code for the variable type.

Bugs:

*******************************************************************************/

int Node::ScanType()
{
	//
	// Don't even try it is we don't know what's happening
	//
	assert(this != NULL);

	//
	// Try to figure out the variable type
	//
	switch (Type)
	{
	case BAS_S_DOUBLE:
		return(VARTYPE_DOUBLE);

	case BAS_S_BYTE:
		return(VARTYPE_BYTE);

	case BAS_S_GFLOAT:
		return(VARTYPE_DOUBLE);

	case BAS_S_HFLOAT:
		return(VARTYPE_DOUBLE);

	case BAS_S_INTEGER:
		return(VARTYPE_INTEGER);

	case BAS_S_LONG:
		return(VARTYPE_LONG);

	case BAS_S_RFA:
		return(VARTYPE_RFA);

	case BAS_S_STRING:
		return(VARTYPE_DYNSTR);

	case BAS_S_WORD:
		return(VARTYPE_WORD);

	case BAS_S_DECIMAL:
		return(VARTYPE_DECIMAL);
	}

	return(0);
}

/*******************************************************************************

Member Function:

	Node::Output()

Title:

	Main function to output translated code.

Description:

	This function outputs the translated code.

Bugs:

	Ugly Code.

*******************************************************************************/

void Node::Output(ostream& os)
{
	//
	// Handle Include Files
	//
	if (NeedIostreamH != 0)
	{
		os << "#include <iostream.h>" << endl;
		NeedIostreamH = 0;
	}

	os << "#include <stdlib.h>" << endl;
	NeedStdlibH = 0;

	if ((NeedStringH != 0) | (NeedBasicFun != 0))
	{
		os << "#include <string.h>" << endl;
		NeedStringH = 0;
	}

	if (NeedMathH != 0)
	{
		os << "#include <math.h>" << endl;
		NeedMathH = 0;
	}

	if (NeedErrorH != 0)
	{
		os << "#include <errno.h>" << endl;
		NeedErrorH = 0;
	}

	if (NeedTimeH != 0)
	{
		os << "#include <time.h>" << endl;
		NeedTimeH = 0;
	}

	os << "#include \"bstring.h\"" << endl;

	if (NeedBasicFun != 0)
	{
		os << "#include \"basicfun.h\"" << endl;
		NeedBasicFun = 0;
	}

	if (NeedDataList != 0)
	{
		os << "#include \"datalist.h\"" << endl;
	}

	os << endl;

	if (NeedRFA != 0)
	{
		os << "typedef char RmsRfa[6];" << endl;
		NeedRFA = 0;
	}

	if (NeedDataList != 0)
	{
		os << "static DataListClass DataList;" << endl;
		NeedDataList = 0;
	}

	os << endl;

	//
	// Output the variables
	//
	Variables->OutputDef(os, Level);

	//
	// Output code
	//
	Level = 0;
	OutputCode(os);
}


/*******************************************************************************

Local Function:

	Node::OutputCode()

Title:

	Output one block of code.

Description:

	This function goes through the level 0 code (statements)
	and handle output of the commands at that level.

Bugs:


*******************************************************************************/

void Node::OutputCode(ostream& os)
{
	assert(this != NULL);

	//
	// Local Variables
	//
	Node *Program = this;

	//
	// Starting at the top node, we scan down handling each
	// level 0 node type as necessary.
	//
	// Could do this by calling OutputCode recursively for the Down
	// level instead of splitting into OutputCode and OutputCodeOne,
	// but that would probibly consume too much stack space.
	//
	while (Program != NULL)
	{
		//
		// Handle Statement. Don't output code from included files.
		//
		if ((Program->FromInclude == 0) || (CompileFlag != 0))
		{
			os << Program;
		}

		//
		// Get next node
		//
		Program = Program->Down;
	}
}


/*******************************************************************************

Local Function:

	Node::OutputBlock()

Title:

	Output one block of code.

Description:

	This function will output a block of code (if it exists)
	surrounded by parentheses, otherwise it will output a semicolon.

Bugs:


*******************************************************************************/

void Node::OutputBlock(ostream& os)
{
	//
	// If this is a null bit of code, just output a semicolon
	//
	if (this == NULL)
	{
		Indent(os);
		os << ";" << endl;
	}
	else
	{
		//
		// Start the block
		//
		Indent(os);
		os << "{" << endl;

		Level++;

		//
		// Local Variables
		//
		Node *Program = this;

		//
		// Starting at the top node, we scan down handling each
		// level 0 node type as necessary.
		//
		// Could do this by calling OutputCode recursively for the Down
		// level instead of splitting into OutputCode and OutputCodeOne,
		// but that would probibly consume too much stack space.
		//
		while (Program != NULL)
		{
			//
			// Handle Statement
			//
			if ((Program->FromInclude == 0) || (CompileFlag != 0))
			{
				os << Program;
			}

			//
			// Get next node
			//
			Program = Program->Down;
		}

		//
		// End the block
		//
		Level--;
		Indent(os);
		os << "}" << endl;
	}
}


/*******************************************************************************

Local Function:

	Node::NoParen()

Title:

	Return a node losing any parentheses around the whole group.

Description:

	This function goes through the level 0 code (statements)
	and handle output of the commands at that level.

Bugs:


*******************************************************************************/

Node& Node::NoParen()
{
	if (Type == '(')
	{
		return *Tree[0];
	}
	else
	{
		return *this;
	}
}

/*******************************************************************************

Local Function:

	Node::OutputCodeOne()

Title:

	Output one block of code.

Description:

	This function goes through the level 0 code (statements)
	and handle output of the commands at that level.

Bugs:


*******************************************************************************/

void Node::OutputCodeOne(ostream& os)
{
	VariableStruct *ThisVar;
	char *ThisPart = 0;

	assert(this != NULL);

	//
	// Handle this node type
	//
	switch (Type)
	{
	//
	// Assignment statement
	//
	case '#':
		os << "BasicChannel[" << Tree[0] << "]";
		break;

	case '=':
		Tree[0]->OutputBstringText(os);
		os << " == " << Tree[1];
		break;

	case '+':
	case '-':
	case '*':
	case '/':
	case '>':
	case '<':
		Tree[0]->OutputBstringText(os);
		os << " " << TextValue << " " << Tree[1];
		break;

	case '(':
		os << "(" << Tree[0] << ")";
		break;

	case '^':
		os << "pow(" << Tree[0]->NoParen() << ", " <<
			Tree[1]->NoParen() << ")";
		break;

	case BAS_S_AND:
		os << Tree[0] << " & " << Tree[1];
		break;

	case BAS_N_ASSIGN:
		Indent(os);
		if (Tree[0] != NULL)
		{
			os << Tree[0];
		}
		else
		{
			os << "(Return)";
		}
		os << " = " << Tree[1] << ";" << endl;
		break;

	case BAS_N_ASSIGNLIST:
		assert(Tree[0] != NULL);
		assert(Tree[1] != NULL);

		os << Tree[0] << " = " << Tree[1];
		break;

	case BAS_S_BY:
		switch(Tree[1]->Type)
		{
		case BAS_S_DESC:
			os << "$descr(" << Tree[0] << ")";
			break;
		case BAS_S_REF:
			os << "&" << Tree[0];
			break;
		default:
			os << Tree[0];
			break;
		}
		break;

	case BAS_S_CALL:
		Indent(os);
		os << Tree[0] << ";" << endl;
		break;

	case BAS_S_CASE:
		if (Tree[0] != NULL)
		{
			Tree[0]->OutputCaseLabel(os);
			os << endl;
		}
		else
		{
			Indent(os);
			os << "case:\t// Empty case label" <<
				endl << endl;
		}

		//
		// Output any info under case statement
		//
		if (Tree[3] != NULL)
		{
			Tree[3]->OutputCode(os);
		}
		Indent(os);
		os << "break;" << endl << endl;
		break;

	case BAS_N_CASEELSE:
		Level--;
		Indent(os);
		os << "default:" << endl;
		Level++;

		//
		// Output any info under case statement
		//
		if (Tree[3] != NULL)
		{
			Tree[3]->OutputCode(os);
		}
		Indent(os);
		os << "break;" << endl << endl;
		break;

	case BAS_P_CDD:
		os << "#dictionary " << Tree[0] << endl;
		break;

	case BAS_S_CHAIN:
		Indent(os);
		os << "BasicChain(" << Tree[0]->NoParen() << ");" << endl;
		break;

	case BAS_S_CLOSE:
		Indent(os);
		os << "BasicChannel[" << Tree[0]->NoParen() <<
			"].Close();" << endl;
		break;

	case BAS_S_COMMON:
		OutputMap(os);
		break;

	case BAS_S_DATA:
		OutputData(os);
		break;

	case BAS_S_DECLARE:
		Tree[2]->OutputDefinition(os, Tree[1], Type);
		break;

	case BAS_S_DEF:

		//
		// Empty variable list
		//
		assert(Tree[2] != 0);

		//
		// Scan the call parameters
		//
		this->VariableScanOne(1);
		Variables->Fixup();

		//
		// Dump out variable names if user asked for it
		//
		if (VariableDump)
		{
			Variables->Dump();
		}

		//
		// Output function name
		//
		os << endl;
		Tree[2]->OutputDefinition(os, Tree[0], Type);

		//
		// Braces starting function
		//
		Indent(os);
		os << "{" << endl;

		//
		// Output local variables
		//
		if (Tree[3] != 0)
		{
			Level++;
			Variables->OutputDef(os, Level);
			os << endl;
			Tree[3]->OutputCode(os);
			Level--;
		}

		Indent(os);
		os << "}" << endl;

		break;

	case BAS_S_DELETE:
		Indent(os);
		os << "BasicChannel[" << Tree[0]->NoParen() <<
			"].Delete();" << endl;
		break;

	case BAS_S_DIM:
		assert(Tree[0] != NULL);
		Tree[0]->OutputDefinition(os, this, Type);
		break;

	case BAS_S_ELSE:
		break;

	case BAS_S_END:
	case BAS_S_FNEND:
		break;

	case BAS_S_EQV:
	case BAS_S_IMP:
		os << "Basic" << TextValue << "(" << Tree[0] << ", " <<
			Tree[1] << ")";
		break;

	case BAS_S_ERL:
		os << "errline";
		break;

	case BAS_S_ERR:
		os << "errno";
		break;

	case BAS_S_EXTERNAL:
		Tree[2]->OutputDefinition(os, Tree[1], Type);
		break;

	case BAS_N_EXTERNALCONSTANT:
		Tree[2]->OutputDefinition(os, Tree[1], Type);
		break;

	case BAS_N_EXTERNALFUNCTION:
		Tree[2]->OutputDefinition(os, Tree[1], Type);
		break;

	case BAS_N_EXTERNALSUB:
		Tree[2]->OutputDefinition(os, Tree[1], Type);
		break;

	case BAS_S_EXIT:
		Indent(os);
		os <<  "return;" << endl;
		break;

	case BAS_S_FIND:
		Indent(os);
		os << "BasicChannel[" << Tree[0]->NoParen() << "].Find(";
		if (Tree[1] != NULL)
		{
			os << Tree[1]->NoParen();
		}
		os << ");" << endl;
		break;

	case BAS_S_FOR:
		Indent(os);
		os << "for (";

		//
		// from
		//
		os << Tree[0] << "; ";

		//
		// to
		//
		os << Tree[0]->Tree[0] << " <= " << Tree[1] << "; ";

		//
		// step
		//
		if (Tree[2] != NULL)
		{
			os << Tree[0]->Tree[0] << " += " << Tree[2];
		}
		else
		{
			os << Tree[0]->Tree[0] << "++";
		}

		os << ")" << endl;

		Tree[3]->OutputBlock(os);

		break;

	case BAS_N_FORASSIGN:
		os << Tree[0] << " = " << Tree[1];
		break;

	case BAS_N_FORUNTIL:
		Indent(os);

		//
		// from
		//
		os << "for (" << Tree[0] << "; ";

		//
		// until
		//
		os << "!(" << Tree[0]->NoParen() << ")" << "; ";

		//
		// step
		//
		os << Tree[0]->Tree[0] << "++)" << endl;

		Tree[3]->OutputBlock(os);

		break;

	case BAS_N_FORWHILE:
		Indent(os);

		//
		// from
		//
		os << "for (" << Tree[0] << "; ";

		//
		// until
		//
		os << Tree[0]->Tree[0] << "==" << Tree[1] << "; ";

		//
		// step
		//
		os << Tree[0]->Tree[0] << "++)" << endl;

		Tree[3]->OutputBlock(os);

		break;

	case BAS_S_FUNCTION:

		//
		// Empty variable list
		//
		assert(Tree[2] != 0);
		Variables->ClearLocal();

		//
		// Scan the call parameters
		//
		this->VariableScanOne(1);
		Variables->Fixup();

		//
		// Dump out variable names if user asked for it
		//
		if (VariableDump)
		{
			Variables->Dump();
		}

		//
		// Output function name
		//
		os << endl;
		Tree[2]->OutputDefinition(os, Tree[0], Type);
		os << endl;

		//
		// Braces starting function
		//
		Indent(os);
		os << "{" << endl;

		//
		// Output local variables
		//
		if (Tree[3] != 0)
		{
			Level++;
			Variables->OutputDef(os, Level);
			os << endl;
			Tree[3]->OutputCode(os);
			Level--;
		}

		Indent(os);
		os << "}" << endl;

		break;

	case BAS_V_FUNCTION:
		//
		// Look for variable in table
		//
		ThisVar = Variables->Lookup(TextValue, Tree[0]);

		//
		// Print this variable
		//
		if (ThisVar != NULL)
		{
			os << ThisVar->GetName();
		}
		else
		{
			os << TextValue;
		}

		if (Tree[0] != NULL)
		{
			os << "(" << Tree[0]->NoParen() << ")";
		}
		else
		{
			os << "()";
		}
		break;

	case BAS_S_FUNCTIONEXIT:
		Indent(os);
		os << "return;" << endl;
		break;

	case BAS_X_GE:
		os << Tree[0] << " >= " << Tree[1];
		break;

	case BAS_S_GET:
		Indent(os);
		os << "BasicChannel[" << Tree[0]->NoParen() << "].Get(";
		if (Tree[1] != NULL)
		{
			os << Tree[1]->NoParen();
		}
		os << ");" << endl;
		break;

	case BAS_S_GOSUB:
		//
		// Yes, I know, C doesn't have a GOSUB, but this puts
		// out code that can be manually edited. Without this,
		// there is currently no way to find the subroutines.
		//
		Indent(os);
		os << "gosub " << Tree[0] << ";" << endl;
		break;

	case BAS_S_GOTO:
		Indent(os);
		os << "goto " << Tree[0] << ";" << endl;
		break;

	case BAS_S_IF:
		Indent(os);
		assert(Tree[0] != NULL);

		os << "if (" << Tree[0]->NoParen() << ")" << endl;

		Tree[3]->OutputBlock(os);

		if (Tree[2] != NULL)
		{
			Indent(os);
			os << "else" << endl;

			Tree[2]->OutputBlock(os);
		}
		break;

	case BAS_P_INCLUDE:
		//
		// Don't output '#include' if we are also outputting the
		// code that was included.
		//
		if (CompileFlag == 0)
		{
			os << "#include " << Tree[0] << endl;
		}
		else
		{
			os << "// #include " << Tree[0] << endl;
		}

		break;

	case BAS_S_INPUT:
		OutputInput(os);
		break;

	case BAS_V_INT:
		if (TextValue[strlen(TextValue) - 1] == '%')
		{
			TextValue[strlen(TextValue) - 1] = '\0';
		}
		os << TextValue;
		break;

	case BAS_V_INTEGER:
	case BAS_V_FLOAT:
		os << TextValue;
		break;

	case BAS_S_ITERATE:
		Indent(os);
		os << "continue;" << endl;
		break;

	case BAS_S_KILL:
		Indent(os);
		os << "unlink(" << Tree[0]->NoParen() << ")" << endl;
		break;

	case BAS_V_LABEL:
		ThisVar = Variables->Lookup(TextValue, 0);
		if (ThisVar != NULL)
		{
			if (ThisVar->EverUsed != 0)
			{
				os << ThisVar->GetName() << ":;" << endl;
			}
		}
		else
		{
			os << TextValue << ":;" << endl;
		}
		break;

	case BAS_X_LE:
		os << Tree[0] << " <= " << Tree[1];
		break;

	case BAS_S_LET:
		Indent(os);
		os << Tree[0] << ";" << endl;
		break;

	case BAS_S_LINPUT:
		Indent(os);
		if (Tree[0] != NULL)
		{
			os << "fscanf(" << Tree[0] << ");" << endl;
		}
		break;

	case BAS_N_LIST:
		if (Tree[0] == NULL)
		{
			os << "0";
		}
		else
		{
			os << Tree[0];
		}

		os << ", ";

		if (Tree[1] == NULL)
		{
			os << "0";
		}
		else
		{
			os << Tree[1];
		}
		break;

	case BAS_S_LSET:
		assert(Tree[0] != NULL);
		assert(Tree[0]->Type == BAS_N_ASSIGN);
		assert(Tree[0]->Tree[0] != NULL);
		assert(Tree[0]->Tree[1] != NULL);

		Indent(os);
		os << "BasicLset(" << Tree[0]->Tree[0] << ", " <<
			Tree[0]->Tree[1] << ");" << endl;
		break;

	case BAS_S_MAP:
		OutputMap(os);
		break;

	case BAS_V_NAME:
		OutputVarName(os);
		break;

	case BAS_X_NEQ:
		os << Tree[0] << " != " << Tree[1];
		break;

	case BAS_S_NOT:
		os << "!" << Tree[0];
		break;

	case BAS_N_NULL:
		break;

	case BAS_N_ONGOTO:
		Indent(os);
		os << "switch(" << Tree[0]->NoParen() << ")" << endl;
		Indent(os);
		os << "{" << endl;
		Tree[1]->OutputOngo(os, "goto", 1);
		Indent(os);
		os << "}" << endl;
		break;

	case BAS_N_ONGOSUB:
		Indent(os);
		os << "switch(" << Tree[0]->NoParen() << ")" << endl;
		Indent(os);
		os << "{" << endl;
		Tree[1]->OutputOngo(os, "Gosub", 1);
		Indent(os);
		os << "}" << endl;
		break;

	case BAS_S_OPEN:
		//
		// For INPUT/OUTPUT
		//
		if (Tree[1] != NULL)
		{
			Indent(os);
			os << "BasicChannel[" << Tree[2] << "].";
			if (Tree[1]->Type == BAS_S_INPUT)
			{
				os << "ForInput();" << endl;
			}
			else
			{
				os << "ForOutput();" << endl;
			}

		}

		//
		// Other flags set on open
		//
		if (Tree[3] != NULL)
		{
			Tree[3]->OutputOpenStuff(os, Tree[2]);
		}


		Indent(os);
		os << "BasicChannel[" << Tree[2] << "].Open(";
		if (Tree[0] != NULL)
		{
			os << Tree[0];
		}
		os << ");" << endl;
		break;

	case BAS_S_OPTION:
		break;

	case BAS_S_OR:
		os << Tree[0] << " | " << Tree[1];
		break;

	case BAS_V_PREDEF:
		os << TextValue;
		break;

	case BAS_S_PRINT:
		OutputPrint(os);
		break;

	case BAS_S_PUT:
		Indent(os);
		os << "BasicChannel[" << Tree[0]->NoParen() <<
			"].Put();" << endl;
		break;

	case BAS_S_RANDOM:
		Indent(os);
		os << "srand(time());" << endl;
		break;

	case BAS_S_RECORD:
		Indent(os);
		os << "struct ";
		if (Tree[0] != NULL)
		{
			os << Tree[0];
		}
		os << endl;
		Indent(os);
		os << "{" << endl;

		if (Tree[3] != NULL)
		{
			Level++;
			Tree[3]->OutputRecord(os);
			Level--;
		}

		Indent(os);
		os << "};" << endl;

		break;

	case BAS_S_RESTORE:
		if (Tree[0] == NULL)
		{
			Indent(os);
			os << "DataList.Reset();" << endl;
		}
		else
		{
			Indent(os);
			os << "BasicChannel[" << Tree[0]->NoParen() <<
				"].Reset();" << endl;
		}
		break;

	case BAS_S_READ:
		Indent(os);
		os << "DataList.Read(" << Tree[0] << ");" << endl;
		break;

	case BAS_S_RSET:
		assert(Tree[0] != NULL);
		assert(Tree[0]->Type == BAS_N_ASSIGN);
		assert(Tree[0]->Tree[0] != NULL);
		assert(Tree[0]->Tree[1] != NULL);

		Indent(os);
		os << "BasicRset(" << Tree[0]->Tree[0] << ", " <<
			Tree[0]->Tree[1] << ");" << endl;
		break;

	case BAS_S_RESET:
		if (Tree[0] == NULL)
		{
			Indent(os);
			os << "DataList.Reset();" << endl;
		}
		else
		{
			Indent(os);
			os << "RmsReset(" << Tree[0]->NoParen() << ");" << endl;
		}
		break;

	case BAS_S_RESUME:
		Indent(os);
		if (Tree[0] != NULL)
		{
			os << "goto " << Tree[0] << ";" << endl;
		}
		else
		{
			os << "BasicResume;" << endl;
		}
		break;

	case BAS_S_RETURN:
		Indent(os);
		os << "GosubReturn;" << endl << endl;
		break;

	case BAS_S_SELECT:
		assert(Tree[0] != NULL);

		Indent(os);
		os << "switch (" << Tree[0]->NoParen() << ")" << endl;

		Tree[3]->OutputBlock(os);

		break;

	case BAS_S_STOP:
		Indent(os);
		os << "exit(0);" << endl;
		break;

	case BAS_X_STRREF:
		os << Tree[0] << "." << Tree[1];
		break;

	case BAS_N_STRUCTNAME:
		os << TextValue;
		break;

	case BAS_S_SUB:

		//
		// Empty variable list
		//
		assert(Tree[2] != 0);
		Variables->ClearLocal();

		Tree[2]->VariableScanOne(1);
		Variables->Fixup();

		//
		// Output function name
		//
		os << endl;

		Tree[2]->OutputDefinition(os, Tree[0], Type);

		//
		// Braces starting function
		//
		Indent(os);
		os << "{" << endl;

		//
		// Output local variables
		//
		if (Tree[3] != 0)
		{
			Level++;

			Tree[3]->VariableScan(0);
			Variables->Fixup();

			Variables->OutputDef(os, Level);
			os << endl;
			Tree[3]->OutputCode(os);
			Level--;
		}

		Indent(os);
		os << "}" << endl;

		break;

	case BAS_S_SUBEXIT:
		Indent(os);
		os <<  "return;" << endl;
		break;

	case BAS_V_TEXTSTRING:
		ThisPart = mangle_string(TextValue);
		os << ThisPart;
		delete ThisPart;
		break;

	case BAS_P_TITLE:
	case BAS_P_SBTTL:
	case BAS_P_IDENT:
	case BAS_S_REMARK:
		Indent(os);
		OutputRemark(os);
		break;

	case BAS_S_TO:
		//
		// TO used in a case statement
		//
		os << Tree[0] << " ... " << Tree[1];
		break;

	case BAS_N_TYPEFUN:
		os << "(";
		OutputDatatype(os, Tree[1]->Type);
		os << ")(" << Tree[0] << ")";
		break;

	case BAS_N_UMINUS:
		os << "-" << Tree[0];
		break;

	case BAS_N_UPLUS:
		os << "+" << Tree[0];
		break;

	case BAS_X_UNARYGT:
		os << "> " << Tree[0];
		break;

	case BAS_X_UNARYGE:
		os << ">= " << Tree[0];
		break;

	case BAS_X_UNARYLE:
		os << "<= " << Tree[0];
		break;

	case BAS_X_UNARYLT:
		os << "< " << Tree[0];
		break;

	case BAS_X_UNARYNEQ:
		os << "!= " << Tree[0];
		break;

	case BAS_S_UNLESS:
		assert(Tree[0] != NULL);

		Indent(os);
		os << "if (!(" << Tree[0]->NoParen() << "))" << endl;

		Tree[3]->OutputBlock(os);

		break;

	case BAS_S_UNLOCK:
		Indent(os);
		os << "BasicChannel[" << Tree[0]->NoParen() <<
			"].Unlock();" << endl;
		break;

	case BAS_S_UNTIL:
		assert(Tree[0] != NULL);

		Indent(os);
		os << "while (!(" << Tree[0]->NoParen() << "))" << endl;

		Tree[3]->OutputBlock(os);

		break;

	case BAS_S_UPDATE:
		Indent(os);
		os << "RmsUpdate(" << Tree[0]->NoParen() << ");" << endl;
		break;

	case BAS_V_USELABEL:
		OutputLabel(os);
		break;

	case BAS_N_VARTYPE:
	case BAS_S_BYTE:
	case BAS_S_DOUBLE:
	case BAS_S_GFLOAT:
	case BAS_S_HFLOAT:
	case BAS_S_LONG:
	case BAS_S_WORD:
	case BAS_S_INTEGER:
	case BAS_S_REAL:
	case BAS_S_RFA:
	case BAS_S_STRING:
	case BAS_S_DECIMAL:

		if (Tree[2] != NULL)
		{
			Tree[2]->OutputDefinition(os, this, Type);
		}
		else
		{
			OutputDatatype(os, Type);
		}
		break;

	case BAS_S_WAIT:
	case BAS_S_SLEEP:
		Indent(os);
		os << "sleep(" << Tree[0]->NoParen() << ");" << endl;
		break;

	case BAS_S_WHILE:
		assert(Tree[0] != NULL);

		Indent(os);
		os << "while (" << Tree[0]->NoParen() << ")" << endl;

		Tree[3]->OutputBlock(os);
		break;

	case BAS_S_XOR:
		os << Tree[0] << " ^ " << Tree[1];
		break;

	default:
		os << "??" << Type << ":" << TextValue << "(";
		if (Tree[0] != NULL)
		{
			os << Tree[0];
		}
		os << ",";
		if (Tree[1] != NULL)
		{
			os << Tree[1];
		}
		os << ",";
		if (Tree[2] != NULL)
		{
			os << Tree[2];
		}
		os << ",";
		if (Tree[3] != NULL)
		{
			os << Tree[3];
		}
		os << ")" << endl;
		break;
	}
}


/*******************************************************************************

Member Function:

	Node::OutputVarName()

Title:

	Outputs Variable Name

Description:

	Creates the full name for a variable

*******************************************************************************/

void Node::OutputVarName(ostream& os)
{
	//
	// Local Variables
	//
	VariableStruct *ThisVar;

	assert(this != NULL);
	assert(TextValue != NULL);

	//
	// Look for variable in table
	//
	ThisVar = Variables->Lookup(TextValue, Tree[0]);

	//
	// Make up a good name to print for this variable
	//
	if (ThisVar != NULL)
	{
		os << ThisVar->GetName();
	}
	else
	{
		os << TextValue;
	}

	//
	// Handle array references (May need to be modified from x[a,b]
	// to x[a][b] manually.
	//
	if (Tree[0] != NULL)
	{
		if ((ThisVar != NULL) && (ThisVar->Class == VARCLASS_FUNC))
		{
			os << "(" << Tree[0]->NoParen() << ")";
		}
		else
		{
			os << "[";
			Tree[0]->OutputArrayParam(os);
			os << "]";
		}
	}
}

/*******************************************************************************

Member Function:

	Node::OutputPrint()

Title:

	Outputs a print statement.

Description:

	This function will mangle a print statement.

Bugs:


*******************************************************************************/

void Node::OutputPrint(ostream& os)
{
	Node *PUsing = 0;
	Node *PChannel = 0;
	Node *TreeDown;

	assert(this != NULL);

	//
	// Indent as necessary
	//
	Indent(os);

	os << "cout";

	if (Tree[0] != NULL)
	{
		if (Tree[0]->OutputIPData(os, 0) != 0)
		{
			os << " << endl";
		}
	}
	else
	{
		os << " << endl";
	}

	os << ";" << endl;
}

/*******************************************************************************

Member Function:

	Node::OutputInput()

Title:

	Outputs a input statement.

Description:

	This function will mangle a input statement.

Bugs:


*******************************************************************************/

void Node::OutputInput(ostream& os)
{
	Node *PChannel = 0;
	Node *TreeDown;

	assert(this != NULL);

	//
	// Indent as necessary
	//
	Indent(os);

	os << "cin";

	if (Tree[0] != NULL)
	{
		Tree[0]->OutputIPData(os, 1);
	}
	os << ";" << endl;
}

/*******************************************************************************

Member Function:

	int Node::OutputIPData()

Title:

	Outputs data related to an input or print statement.

Description:

	This function will mangle a input statement. Returns a
	flag depending on if the input ended in something that should
	cause a return to be appended.

Parameters:

	InputFlag

		Specifies if it is working on an input statement,
		ie. specifies the direction of the arrows.

Return:

	0 - Blank item on right hand side.
	1 - Normal item on right hand side.
	2 - Special item was just output

Bugs:


*******************************************************************************/

int Node::OutputIPData(ostream& os, int InputFlag)
{
	int ReturnFlag;		// Does a return seem to be needed
				//   Gets set to one if the print/Input command
				//   ends with a ',' or ';'.
	int TestFlag = 0;	// Test for special item

	assert(this != NULL);

	//
	// Is this a piece of a list?
	//
	switch (Type)
	{
	case BAS_N_LIST:
		//
		// Try the left side of the tree
		//
		if (Tree[0] != NULL)
		{
			TestFlag = Tree[0]->OutputIPData(os, InputFlag);
		}

		//
		// Do we need to append a tab between fields?
		// Only do so for print statements.
		//
		if ((InputFlag == 0) && (TextValue[0] == ',') &&
			(TestFlag != 2))
		{
			os << " << '\\t'";
		}

		//
		// Handle right side of the tree
		//
		if (Tree[1] != NULL)
		{
			ReturnFlag = Tree[1]->OutputIPData(os, InputFlag);
		}
		else
		{
			ReturnFlag = 0;
		}

		break;

	case '#':
		os << " << BasicChannel[" << Tree[0] << "]";
		ReturnFlag = 2;
		break;

	case BAS_S_USING:
		os << " << USING(" << Tree[0] << ")";
		ReturnFlag = 2;
		break;

	case BAS_N_RECORD:
		os << " << RECORD(" << Tree[0] << ")";
		ReturnFlag = 2;
		break;

	default:
		//
		// A quoted string is always output, even in input
		// statements.
		//
		if ((InputFlag == 0) || (Type == BAS_V_TEXTSTRING))
		{
			os << " << ";
		}
		else
		{
			os << " >> ";
		}

		os << this;

		ReturnFlag = 1;

		break;
	}

	//
	// Done
	//
	return ReturnFlag;
}

/*******************************************************************************

Global Function:

	Node::OutputData()

Title:

	Outputs a data statement.

Description:

	This function will mangle a data statement.

Bugs:


*******************************************************************************/

void Node::OutputData(ostream& os)
{
	assert(this != NULL);

	Indent(os);
	os << "DataList.Append(";
	Tree[0]->OutputDataValue(os);
	os << ");" << endl;
}


/*******************************************************************************

Local Function:

	Node::OutputDataValue()

Title:

	Outputs one data value in a data statement.

Description:

	This function will mangle one data value for a data statement.
	It forces everything to be a string.

Bugs:


*******************************************************************************/

void Node::OutputDataValue(ostream& os)
{
	assert(this != NULL);

	if (Type == BAS_N_LIST)
	{
		Tree[0]->OutputDataValue(os);
		os << ",";
		Tree[1]->OutputDataValue(os);
	}
	else
	{
		if (Type == BAS_V_TEXTSTRING)
		{
			os << this;
		}
		else
		{
			//
			// Force quotes ariund data value
			//
			os << '"' << this  << '"';
		}
	}
}

/*******************************************************************************

Global Function:

	Node::OutputMap()

Title:

	Outputs a map or common statement.

Description:

	This function will mangle a map or common statement.

Bugs:


*******************************************************************************/

void Node::OutputMap(ostream& os)
{
	assert(this != NULL);
	assert(Tree[0] != NULL);

	VariableStruct* ThisVar = Variables->Lookup(Tree[0]->TextValue, NULL);
	assert(ThisVar != NULL);

	Indent(os);
	os << "extern struct" << endl;

	Indent(os);
	os << "{" << endl;

	Level++;
	if (Tree[1] != NULL)
	{
		Tree[1]->OutputOneMap(os, ThisVar);
	}
	Level--;
	Indent(os);
	os << "} " << ThisVar->GetName() << ";" << endl;
}

/*******************************************************************************

Member Function:

	Node::OutputOneMap()

Title:

	Outputs a map or common statement.

Description:

	This function will mangle a map or common statement.

Bugs:


*******************************************************************************/


void Node::OutputOneMap(ostream& os, VariableStruct* MapVar)
{
	assert(this != NULL);

	switch (Type)
	{
	case BAS_N_LIST:

		if (Tree[0] != NULL)
		{
			Tree[0]->OutputOneMap(os, MapVar);
		}
		if (Tree[1] != NULL)
		{
			Tree[1]->OutputOneMap(os, MapVar);
		}
		break;

	case BAS_V_NAME:

		Indent(os);
		os << TextValue << ";" << endl;

		break;

	case BAS_N_VARTYPE:
	case BAS_S_STRING:
	case BAS_S_BYTE:
	case BAS_S_DOUBLE:
	case BAS_S_GFLOAT:
	case BAS_S_HFLOAT:
	case BAS_S_LONG:
	case BAS_S_WORD:
	case BAS_S_INTEGER:
	case BAS_S_REAL:
	case BAS_S_RFA:
	case BAS_S_DECIMAL:

		assert(Tree[0] != NULL);

		Indent(os);
		os << TextValue << " " << Tree[0]->TextValue;
		os << ";" << endl;
		break;

	default:
		Indent(os);
		OutputVarName(os);
		os << ";;;" << endl;
		break;
	}

	if (Down != NULL)
	{
		Down->OutputOneMap(os, MapVar);
	}
}

/*******************************************************************************

Global Function:

	Node::ScanMap()

Title:

	Outputs a map or common statement.

Description:

	This function will mangle a map or common statement.

Bugs:


*******************************************************************************/

void Node::ScanMap()
{
	assert(this != NULL);
	assert(Tree[0] != NULL);

	VariableStruct* ThisVar = new VariableStruct(Tree[0]->TextValue,
		VARTYPE_STRUCT, VARCLASS_NONE, TRUE);
	Variables->Append(ThisVar);
	Variables->Fixup();

	if (Tree[1] != NULL)
	{
		Tree[1]->ScanOneMap(ThisVar);
	}
}

/*******************************************************************************

Member Function:

	Node::ScanOneMap()

Title:

	Outputs a map or common statement.

Description:

	This function will mangle a map or common statement.

Bugs:


*******************************************************************************/


void Node::ScanOneMap(VariableStruct* MapVar)
{
	char NameBuf[128];
	char TempBuf[64];

	VariableStruct* ThisVar;

	assert(this != NULL);

	switch (Type)
	{
	case BAS_N_LIST:

		if (Tree[0] != NULL)
		{
			Tree[0]->ScanOneMap(MapVar);
		}
		if (Tree[1] != NULL)
		{
			Tree[1]->ScanOneMap(MapVar);
		}
		break;

	case BAS_V_NAME:

		ThisVar = new VariableStruct(TextValue,
			0, VARCLASS_NONE, TRUE);
		Variables->Append(ThisVar);

		strcpy(TempBuf, TextValue);
		strlwr(TempBuf);
		sprintf(NameBuf, "%s.%s", MapVar->GetName(), TempBuf);
		ThisVar->SetCName(NameBuf);

		break;

	case BAS_N_VARTYPE:
	case BAS_S_BYTE:
	case BAS_S_DOUBLE:
	case BAS_S_GFLOAT:
	case BAS_S_HFLOAT:
	case BAS_S_LONG:
	case BAS_S_WORD:
	case BAS_S_INTEGER:
	case BAS_S_REAL:
	case BAS_S_RFA:
	case BAS_S_STRING:
	case BAS_S_DECIMAL:

		assert(Tree[0] != NULL);

		ThisVar = new VariableStruct(Tree[0]->TextValue,
			0, VARCLASS_NONE, TRUE);
		Variables->Append(ThisVar);

		strcpy(TempBuf, Tree[0]->TextValue);
		strlwr(TempBuf);
		sprintf(NameBuf, "%s.%s",
			MapVar->GetName(), TempBuf);
		ThisVar->SetCName(NameBuf);

		break;

	default:
		cerr << "Undefined type "<< Type << endl;
		break;
	}

	if (Down != NULL)
	{
		Down->ScanOneMap(MapVar);
	}
}

/*******************************************************************************

Member Function:

	Node::OutputCaseLabel()

Title:

	Output the "case" selections, one at a time.

Description:


Bugs:


*******************************************************************************/

void Node::OutputCaseLabel(ostream& os)
{
	assert(this != NULL);

	//
	// Handle this if it is a list
	//
	if (Type == BAS_N_LIST)
	{
		Tree[0]->OutputCaseLabel(os);
		Tree[1]->OutputCaseLabel(os);
		return;
	}

	//
	// Output one case statement
	//
	Level--;
	Indent(os);
	os << "case " << this << ":" << endl;
	Level++;
}

/*******************************************************************************

Local Function:

	Indent(os)

Title:

	Indents to current level.

Description:

	This function is used to indent the code.

Bugs:


*******************************************************************************/

void Node::Indent(ostream& os)
{
	int TabLoop;

	for (TabLoop = 1; TabLoop <= Level; TabLoop++)
	{
		os << "\t";
	}
}

/*******************************************************************************

Local Function:

	OutputRemark()

Title:

	Prints out a comment.

Description:

	This function strips off all of the junk from a comment
	and prints it out.

Bugs:


*******************************************************************************/

void Node::OutputRemark(ostream& os)
{
	char* Remark = TextValue;
	int pointer;

	assert(TextValue != NULL);

	//
	// Skip start of remark character
	//
	if (*Remark == '!')
	{
		Remark++;
	}

	//
	// Skip extra white space
	//
	if (*Remark == ' ')
	{
		Remark++;
	}

	//
	// Lose any junk at end of comment
	// (Yes, it does modify the original value)
	//
	while((pointer = strlen(Remark) - 1) >= 0)
	{
		// Line continuation marker
		if (Remark[pointer] == '&')
		{
			Remark[pointer] = '\0';
			continue;
		}

		// New Line
		if (Remark[pointer] == '\n')
		{
			Remark[pointer] = '\0';
			continue;
		}

		// Space
		if (Remark[pointer] == ' ')
		{
			Remark[pointer] = '\0';
			continue;
		}

		// Tab
		if (Remark[pointer] == '\t')
		{
			Remark[pointer] = '\0';
			continue;
		}

		// Exclimation point
		if (Remark[pointer] == '!')
		{
			Remark[pointer] = '\0';
			continue;
		}

		// Break out of loop
		break;
	}

	//
	// Dump out remark (and maybe add a space to the front)
	//
	if ((*Remark == '\0') || (*Remark == '\t') || (*Remark == '*'))
	{
		os << "//" << Remark;
	}
	else
	{
		os << "// " << Remark;
	}

	//
	// There may be some extra stuff that needs printing
	//
	if (Tree[0] != NULL)
	{
		os << ' ';
		Tree[0]->OutputCode(os);
	}

	os << endl;

}

/*******************************************************************************

Function:

	OutputDatatype(os, int Type)

Title:

	Prints out a data type (in C++).

Description:

	Converts a BASIC data type into a C++ Data type
Bugs:


*******************************************************************************/

void Node::OutputDatatype(ostream& os, int ThisType)
{
	char* DataType = "";

	switch(ThisType)
	{
	case VARTYPE_VOID:
		DataType = "void";
		break;

	case VARTYPE_LABEL:
		DataType = "label";	// ?? Shouldn't ever see
		break;

	case BAS_S_BYTE:
	case VARTYPE_BYTE:
		DataType = "char";
		break;

	case BAS_S_DOUBLE:
	case VARTYPE_DOUBLE:
		DataType = "double";
		break;

	case BAS_S_GFLOAT:
		DataType = "double /* GFloat */";
		break;

	case BAS_S_HFLOAT:
		DataType = "double /* HFloat */";
		break;

	case BAS_S_LONG:
	case VARTYPE_LONG:
		DataType = "long";
		break;

	case BAS_S_WORD:
	case VARTYPE_WORD:
		DataType = "short int";
		break;

	case BAS_S_INTEGER:
	case VARTYPE_INTEGER:
		DataType = "int";
		break;

	case BAS_S_REAL:
	case VARTYPE_REAL:
		DataType = "float";
		break;

	case BAS_S_STRING:
	case VARTYPE_DYNSTR:
	case VARTYPE_FIXSTR:
		DataType = "BString";
		break;

	case BAS_S_RFA:
	case VARTYPE_RFA:
		DataType = "RmsRfa";
		break;

	case BAS_S_DECIMAL:
		DataType = "Decimal";
		break;

	default:
		DataType = TextValue;
	}

	os << DataType << " ";
}

/*******************************************************************************

Function:

	OutputOngo()

Title:

	Prints out a computed goto/gosub statement

Description:

	Converts a BASIC data type into a C++ Data type
Bugs:


*******************************************************************************/

void Node::OutputOngo(ostream& os, const char* OnType, int GoLevel)
{
	switch(Type)
	{
	case BAS_N_LIST:
		Tree[0]->OutputOngo(os, OnType, GoLevel);
		Tree[1]->OutputOngo(os, OnType, GoLevel + 1);
		break;

	case BAS_V_USELABEL:
		Indent(os);
		os << "case " << GoLevel << ":" << endl;
		Level++;
		Indent(os);
		os << OnType << ' ';
		OutputLabel(os);
		os << ";" << endl;
		Indent(os);
		os << "break;" << endl;
		Level--;
		break;

	case BAS_S_OTHERWISE:
		Indent(os);
		os << "default:" << endl;
		Level++;
		Indent(os);
		os << OnType << ' ';
		OutputLabel(os);
		os << ";" << endl;
		Indent(os);
		os << "break;" << endl;
		break;

	default:
		os << "Error in computed go (" << this << ")" << endl;
		break;
	}
}

/*******************************************************************************

Function:

	OutputLabel()

Title:

	Prints out a label

Description:

	Output a label (if referenced anywhere)

Bugs:


*******************************************************************************/

void Node::OutputLabel(ostream& os)
{
	VariableStruct *ThisVar = Variables->Lookup(TextValue, 0);
	if (ThisVar != NULL)
	{
		os << ThisVar->GetName();
	}
	else
	{
		os << TextValue;
	}
}

/*******************************************************************************

Local Function:

	mangle_string()

Title:

	Prints out a quoted string.

Description:

	This function processes a quoted string

Bugs:


*******************************************************************************/

static char *mangle_string(char *Text)
{
	char *ThisPart;
	char *ThisPtr;
	char qmark;
	int Base = 10;

	assert(Text != NULL);

	//
	// Create area to work in
	// (Probibly more than we will need)
	//
	ThisPart = new char[strlen(Text) * 2];
	ThisPtr = ThisPart;

	//
	// Handle any possible bases on front of string
	//
	switch (*Text)
	{
	case 'd':
	case 'D':
		Base = 10;
		Text++;
		break;
	case 'b':
	case 'B':
		Base = 2;
		Text++;
		break;
	case 'o':
	case 'O':
		Base = 8;
		Text++;
		break;
	case 'x':
	case 'X':
		Base = 16;
		Text++;
		break;
	}

	//
	// Determine quote character used (Better start with a quote
	// or were in deep ****)
	//
	qmark = *Text++;
	*ThisPtr++ = '"';

	//
	// Process all the characters in original string
	//
	while (*Text)
	{
		switch (*Text)
		{
		//
		// Slash must be doubled
		//
		case '\\':
			*ThisPtr++ = '\\';
			*ThisPtr++ = '\\';
			break;

		//
		// Change (audible bell) character to (\a)
		//
		case '\a':
			*ThisPtr++ = '\\';
			*ThisPtr++ = 'a';
			break;

		//
		// Change (backspace) character to (\b)
		//
		case '\b':
			*ThisPtr++ = '\\';
			*ThisPtr++ = 'b';
			break;

		//
		// Change (tab) character to (\t)
		//
		case '\t':
			*ThisPtr++ = '\\';
			*ThisPtr++ = 't';
			break;

		//
		// Change (vertical tab) character to (\v)
		//
		case '\v':
			*ThisPtr++ = '\\';
			*ThisPtr++ = 'v';
			break;

		//
		// Single quote mark
		//
		case '\'':
			if (qmark == '\'')
			{
				*ThisPtr++ = '"';
				qmark = '\0';
			}
			else
			{
				*ThisPtr++ = '\'';
			}
			break;

		//
		// Double Quote Mark
		//
		case '"':
			if (qmark == '"')
			{
				*ThisPtr++ = '"';
				qmark = '\0';
			}
			else
			{
				*ThisPtr++ = '\\';
				*ThisPtr++ = '"';
			}
			break;

		case 'B':
		case 'b':
		case 'W':
		case 'w':
		case 'L':
		case 'l':
		case 'f':
		case 'F':
		case 'd':
		case 'D':
		case 'g':
		case 'G':
		case 'h':
		case 'H':
		case 'c':
		case 'C':
			if (qmark == 0)
			{
				*ThisPtr = '\0';
				mangle_string_value(ThisPart, Base, *Text);
				ThisPtr = ThisPart + strlen(ThisPart);
			}
			else
			{
				*ThisPtr++ = *Text;
			}
			break;

		default:
			*ThisPtr++ = *Text;
			break;
		}

		Text++;
	}

	//
	// Return result
	//
	*ThisPtr = '\0';
	return (ThisPart);
}

/*******************************************************************************

Local Function:

	mangle_string_value()

Title:

	Prints out a number in a given base.

Description:

	This function processes a quoted string

Bugs:


*******************************************************************************/

static void mangle_string_value(char *ThisPart, int Base, char Type)
{
	char *bascode;
	int len;
	char worktext[64];
	int value;
	int valueflag = 0;

	assert(ThisPart != NULL);

	//
	// Set up format for outputing of this number
	//
	switch(Base)
	{
	case 2:
		bascode = "0b%s";
		break;
	case 8:
		bascode = "0%s";
		break;
	case 16:
		bascode = "0x%s";
		break;
	default: // 10
		bascode = "0x%x";
		valueflag = -1;
		value = atoi(ThisPart + 1);
		break;
	}

	//
	// Lose final quote mark
	//
	len = strlen(ThisPart);
	if (len > 0)
	{
		ThisPart[len-1] = '\0';
	}

	switch(Type)
	{
	case 'B':
	case 'b':
	case 'W':
	case 'w':
		if (valueflag)
		{
			sprintf(worktext, bascode, value);
		}
		else
		{
			sprintf(worktext, bascode, ThisPart+1);
		}
		break;

	case 'L':
	case 'l':
		if (valueflag)
		{
			sprintf(worktext, bascode, value);
		}
		else
		{
			sprintf(worktext, bascode, ThisPart+1);
		}
		strcat(worktext, "L");
		break;

	case 'C':
	case 'c':
		worktext[0] = '\'';
		worktext[1] = '\\';
		if (valueflag)
		{
			sprintf(worktext + 2, bascode, value);
		}
		else
		{
			sprintf(worktext + 2, bascode, ThisPart+1);
		}
		strcat(worktext, "'");
		break;

	default:
		strcpy(worktext, ThisPart+1);
		break;

	}

	strcpy(ThisPart, worktext);
}

/*******************************************************************************

Function:

	OutputDefinition()

Title:

	Prints out a definition

Description:


Bugs:


*******************************************************************************/

void Node::OutputDefinition(ostream& os, Node* MainType, int ExtType)
{
	VariableStruct *ThisVar;

	assert(this != NULL);

	switch(Type)
	{
	case BAS_N_LIST:

		assert(Tree[0] != NULL);
		assert(Tree[1] != NULL);

		//
		// Put multiple items on seperate lines
		//
		Tree[0]->OutputDefinition(os, MainType, ExtType);

		//
		// Do we need a seperator
		//
		switch(ExtType)
		{
		case BAS_S_DECLARE:
		case BAS_S_DIM:
		case BAS_S_EXTERNAL:
		case BAS_N_EXTERNALCONSTANT:
		case BAS_S_SUB:
		case BAS_S_DEF:

			break;

		default:

			os << ", ";
			break;
		}

		Tree[1]->OutputDefinition(os, MainType, ExtType);

		break;

	case BAS_N_ASSIGN:

		Indent(os);
		Tree[0]->OutputDefinition(os, MainType, BAS_N_ASSIGN);
		os << " = " << Tree[1] << ";" << endl;
		break;

	case BAS_N_VARTYPE:

		os << TextValue << " ";
		if (Tree[2] != NULL)
		{
			os << Tree[2];
		}
		if (Tree[3] != NULL)
		{
			os << "[" << Tree[3] << "]";
		}
		break;

	case BAS_S_BYTE:
	case BAS_S_DOUBLE:
	case BAS_S_GFLOAT:
	case BAS_S_HFLOAT:
	case BAS_S_LONG:
	case BAS_S_WORD:
	case BAS_S_INTEGER:
	case BAS_S_REAL:
	case BAS_S_RFA:
	case BAS_S_STRING:
	case BAS_S_DECIMAL:

		//
		// This is a data type, so see if there are any variables
		// listed under this type, else dump out the data type
		// all be itself.
		//
		if (Tree[2] != NULL)
		{
			Tree[2]->OutputDefinition(os, this, ExtType);
		}
		else
		{
			OutputDatatype(os, Type);
		}
		if (Tree[3] != NULL)
		{
			os << "[" << Tree[3] << "]";
		}
		break;

	default:
		//
		// Handle class
		//
		switch(ExtType)
		{
		case BAS_S_EXTERNAL:
		case BAS_N_EXTERNALFUNCTION:
		case BAS_N_EXTERNALSUB:
			Indent(os);
			os << "extern ";
			break;

		case BAS_N_EXTERNALCONSTANT:
			Indent(os);
			os << "extern const ";
			break;

		case BAS_S_DEF:
			Indent(os);
			os << "static ";
			break;

		case BAS_S_DECLARE:
		case BAS_S_DIM:
		case BAS_N_VARTYPE:
		case BAS_S_BYTE:
		case BAS_S_DOUBLE:
		case BAS_S_GFLOAT:
		case BAS_S_HFLOAT:
		case BAS_S_LONG:
		case BAS_S_WORD:
		case BAS_S_INTEGER:
		case BAS_S_REAL:
		case BAS_S_RFA:
		case BAS_S_STRING:
		case BAS_S_RECORD:
		case BAS_S_DECIMAL:

			Indent(os);
			break;

		}

		switch (Type)
		{
		case BAS_V_NAME:
			//
			// Look for variable in table
			//
			ThisVar = Variables->Lookup(TextValue, Tree[0]);

			//
			// Slap on sone kind of data type.
			//	1. Try the variable table
			//	2. Try for any kind of attached type
			//	3. Give up and call it a void.
			//
			if (ThisVar != NULL)
			{
				OutputDatatype(os, ThisVar->Type);
			}
			else
			{
				if (MainType != NULL)
				{
					MainType->OutputDatatype(os,
						MainType->Type);
				}
				else
				{
					os << "void ";
				}
			}

			//
			// Make up a good name to print for this variable
			//
			if (ThisVar != NULL)
			{
				os << ThisVar->GetName();
			}
			else
			{
				os << TextValue;
			}

			//
			// Handle references (Function or array)
			// (May need to be modified from x[a,b]
			// to x[a][b] manually.
			//
			if (Tree[0] != NULL)
			{
				os << "[";
				Tree[0]->OutputArrayParam(os);
				os << "]";
			}
			if (Tree[3] != NULL)
			{
				os << "[" << Tree[3] << "]";
			}
			break;


		case BAS_V_FUNCTION:
		case BAS_S_SUB:

			//
			// Look for variable in table
			//
			ThisVar = Variables->Lookup(TextValue, Tree[0]);

			//
			// Slap on sone kind of data type.
			//	1. Try the variable table
			//	2. Try for any kind of attached type
			//	3. Give up and call it a void.
			//
			if (ThisVar != NULL)
			{
				OutputDatatype(os, ThisVar->Type);
			}
			else
			{
				if (MainType != NULL)
				{
					MainType->OutputDatatype(os,
						MainType->Type);
				}
				else
				{
					os << "void ";
				}
			}

			//
			// Make up a good name to print for this variable
			//
			if (ThisVar != NULL)
			{
				os << ThisVar->GetName();
			}
			else
			{
				os << TextValue;
			}

			//
			// Handle references (Function)
			//
			os << "(";

			if (Tree[0] != NULL)
			{
				Tree[0]->OutputDefinition(os, NULL, BAS_N_LIST);
			}

			os << ")";

			break;

		default:
			os << "(??) " << Type << "(??)" << this;
			break;
		}

		//
		// End definition
		//
		switch(ExtType)
		{
		case BAS_S_DECLARE:
		case BAS_S_DIM:
		case BAS_S_EXTERNAL:
		case BAS_N_EXTERNALFUNCTION:
		case BAS_N_EXTERNALSUB:
		case BAS_N_EXTERNALCONSTANT:
		case BAS_S_RECORD:
		case BAS_N_VARTYPE:
		case BAS_S_BYTE:
		case BAS_S_DOUBLE:
		case BAS_S_GFLOAT:
		case BAS_S_HFLOAT:
		case BAS_S_LONG:
		case BAS_S_WORD:
		case BAS_S_INTEGER:
		case BAS_S_REAL:
		case BAS_S_RFA:
		case BAS_S_STRING:
		case BAS_S_DECIMAL:

			os << ";" << endl;
			break;

		case BAS_V_FUNCTION:
		case BAS_S_SUB:
		case BAS_S_DEF:

			os << endl;
			break;
		}

		break;

	}
}

/*******************************************************************************

Function:

	OutputRecord()

Title:

	Prints out a record structure

Description:


Bugs:


*******************************************************************************/

void Node::OutputRecord(ostream& os)
{
	assert(this != NULL);

	switch(Type)
	{
	case BAS_S_VARIANT:

		Indent(os);
		os << "union" << endl;
		Indent(os);
		os << "{" << endl;
		Level++;
		Tree[3]->OutputRecord(os);
		Level--;
		Indent(os);
		os << "};" << endl;

		break;

	case BAS_S_CASE:
		//
		// Variant case
		//
		Indent(os);
		os << "struct" << endl;
		Indent(os);
		os << "{" << endl;
		Level++;
		Tree[3]->OutputRecord(os);
		Level--;
		Indent(os);
		os << "};" << endl;

		break;

	case BAS_N_VARTYPE:
	case BAS_S_BYTE:
	case BAS_S_DOUBLE:
	case BAS_S_GFLOAT:
	case BAS_S_HFLOAT:
	case BAS_S_LONG:
	case BAS_S_WORD:
	case BAS_S_INTEGER:
	case BAS_S_REAL:
	case BAS_S_RFA:
	case BAS_S_DECIMAL:

		Tree[2]->OutputDefinition(os, this, Type);
		break;

	default:
		OutputCode(os);
		break;
	}

	//
	// Any more levels to structure
	//
	if (Down != NULL)
	{
		Down->OutputRecord(os);
	}
}

/*******************************************************************************

Function:

	OutputArrayParam()

Title:

	Prints out an array reference list

Description:


Bugs:


*******************************************************************************/

void Node::OutputArrayParam(ostream& os)
{
	assert(this != NULL);

	switch(Type)
	{
	case BAS_N_LIST:

		if (Tree[0] != NULL)
		{
			Tree[0]->OutputArrayParam(os);
		}
		else
		{
			os << "(NULL)";
		}
		os << "][";
		if (Tree[1] != NULL)
		{
			Tree[1]->OutputArrayParam(os);
		}
		else
		{
			os << "(NULL)";
		}
		break;

	default:

		os << this;
		break;

	}
}


/*******************************************************************************

Member Function:

	OutputOpenStuff()

Title:

	Output code for an 'OPEN' statement

Description:

	This function generates the code necessary to handle
	a full 'RMS' open statement. It generates calls to a class
	for all the elements available in the statement.

*******************************************************************************/

void Node::OutputOpenStuff(ostream& os, Node* Channel)
{
	switch(Type)
	{
	case BAS_S_RECORDSIZE:

		assert(Tree[0] != NULL);

		Indent(os);
		os << "BasicChannel[" << Channel->NoParen() <<
			"].SetRecordSize(" << Tree[0] << ");" << endl;
		break;

	case BAS_S_FILESIZE:

		assert(Tree[0] != NULL);

		Indent(os);
		os << "BasicChannel[" << Channel->NoParen() <<
			"].SetFileSize(" << Tree[0] << ");" << endl;
		break;

	case BAS_S_EXTENDSIZE:

		assert(Tree[0] != NULL);

		Indent(os);
		os << "BasicChannel[" << Channel->NoParen() <<
			"].SetExtendSize(" << Tree[0] << ");" << endl;
		break;

	case BAS_S_WINDOWSIZE:

		assert(Tree[0] != NULL);

		Indent(os);
		os << "BasicChannel[" << Channel->NoParen() <<
			"].SetWindowSize(" << Tree[0] << ");" << endl;
		break;

	case BAS_S_BLOCKSIZE:

		assert(Tree[0] != NULL);

		Indent(os);
		os << "BasicChannel[" << Channel->NoParen() <<
			"].SetBlockSize(" << Tree[0] << ");" << endl;
		break;

	case BAS_S_BUCKETSIZE:

		assert(Tree[0] != NULL);

		Indent(os);
		os << "BasicChannel[" << Channel->NoParen() <<
			"].SetBucketSize(" << Tree[0] << ");" << endl;
		break;

	case BAS_S_MODE:

		assert(Tree[0] != NULL);

		Indent(os);
		os << "BasicChannel[" << Channel->NoParen() <<
			"].SetMode(" << Tree[0] << ");" << endl;
		break;

	case BAS_S_BUFFER:

		assert(Tree[0] != NULL);

		Indent(os);
		os << "BasicChannel[" << Channel->NoParen() <<
			"].SetBuffer(" << Tree[0] << ");" << endl;
		break;

	case BAS_S_DEFAULTNAME:

		assert(Tree[0] != NULL);

		Indent(os);
		os << "BasicChannel[" << Channel->NoParen() <<
			"].SetDefaultName(" << Tree[0] << ");" << endl;
		break;

	case BAS_S_PRIMARY:

		assert(Tree[0] != NULL);

		Indent(os);
		os << "BasicChannel[" << Channel->NoParen() <<
			"].SetKey(" << Tree[0] << ");" << endl;
		break;

	case BAS_S_ALTERNATE:

		assert(Tree[0] != NULL);

		Indent(os);
		os << "BasicChannel[" << Channel->NoParen() <<
			"].SetKey(" << Tree[0] << ");" << endl;
		break;

	case BAS_S_TEMPORARY:

		Indent(os);
		os << "BasicChannel[" << Channel->NoParen() <<
			"].SetTemporary(True);" << endl;
		break;

	case BAS_S_CONTIGUOUS:

		Indent(os);
		os << "BasicChannel[" << Channel->NoParen() <<
			"].SetContiguous(True);" << endl;
		break;

	case BAS_S_NOREWIND:

		Indent(os);
		os << "BasicChannel[" << Channel->NoParen() <<
			"].SetRewind(False);" << endl;
		break;

	case BAS_S_NOSPAN:

		Indent(os);
		os << "BasicChannel[" << Channel->NoParen() <<
			"].SetSpan(False);" << endl;
		break;

	case BAS_S_SPAN:

		Indent(os);
		os << "BasicChannel[" << Channel->NoParen() <<
			"].SetSpan(True);" << endl;
		break;

	case BAS_S_ORGANIZATION:

		assert(Tree[0] != NULL);

		Indent(os);
		os << "BasicChannel[" << Channel->NoParen() <<
			"].SetOrginization(" << Tree[0]->TextValue;
		if (Tree[0]->Tree[0] != NULL)
		{
			os << ", " << Tree[0]->Tree[0]->TextValue;
		}
		os << ");" << endl;
		break;

	case BAS_S_ACCESS:

		assert(Tree[0] != NULL);

		Indent(os);
		os << "BasicChannel[" << Channel->NoParen() <<
			"].SetAccess(" << Tree[0]->TextValue <<
			");" << endl;
		break;

	case BAS_S_ALLOW:

		assert(Tree[0] != NULL);

		Indent(os);
		os << "BasicChannel[" << Channel->NoParen() <<
			"].SetAllow(" << Tree[0]->TextValue <<
			");" << endl;
		break;

	case BAS_S_CONNECT:

		assert(Tree[0] != NULL);

		Indent(os);
		os << "BasicChannel[" << Channel->NoParen() <<
			"].SetConnect(BasicChannel[" << Tree[0] << "]);" <<
			endl;
		break;

	case BAS_S_MAP:

		assert(Tree[0] != NULL);

		Indent(os);
		os << "BasicChannel[" << Channel->NoParen() <<
			"].SetMap(" << Tree[0] << ", sizeof(" << Tree[0] <<
			"));" << endl;
		break;

	case BAS_S_RECORDTYPE:

		assert(Tree[0] != NULL);

		Indent(os);
		os << "BasicChannel[" << Channel->NoParen() <<
			"].SetRecordType(" << Tree[0]->TextValue <<
			");" << endl;
		break;

	default:

		Indent(os);
		os << "BasicChannel[" << Channel->NoParen() <<
			"].GoofyOpenStuff(" << this << ");" << endl;
		break;

	}

	//
	// Handle downlink
	//
	if (Down != NULL)
	{
		Down->OutputOpenStuff(os, Channel);
	}
}

void Node::OutputBstringText(ostream& os)
{
	switch(this->Type)
	{
	case BAS_V_TEXTSTRING:
		os << "BString(";
		OutputCodeOne(os);
		os << ")";
		break;

	default:
		OutputCodeOne(os);
		break;
	}
}


